/*
 * Author:	William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990, 1991, 1992, William Cheng.
 * 
 * Permission limited to the use, copy, modify, and distribute this software
 * and its documentation for any purpose is hereby granted by the Author without
 * fee, provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the Author not be used
 * in advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The Author makes no
 * representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.  All other
 * rights are reserved by the Author.
 *
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
 * USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */
#ifndef lint
static char RCSid[] =
      "@(#)$Header: /amnt/kona/tangram/u/william/X11/TGIF2/RCS/mainloop.c,v 2.79.1.5 1992/11/03 06:57:40 william Exp $";
#endif

#include <stdio.h>
#include <signal.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "const.h"
#include "patchlevel.h"
#include "types.h"

#include "animate.e"
#include "auxtext.e"
#include "choice.e"
#include "cmd.e"
#include "color.e"
#include "copypaste.e"
#include "cursor.e"
#include "dialog.e"
#include "drawing.e"
#include "file.e"
#include "font.e"
#include "grid.e"
#include "mainmenu.e"
#include "menu.e"
#include "msg.e"
#include "names.e"
#include "obj.e"
#include "raster.e"
#include "ruler.e"
#include "scroll.e"
#include "select.e"
#include "setup.e"
#include "stk.e"
#include "text.e"
#include "version.e"
#include "xbitmap.e"
#include "xpixmap.e"

#ifdef USE_XAPPLRESDIR
extern char	* getenv ();
extern int	putenv ();
#endif

int	cmdLineBW = FALSE;
int	cmdLineRV = INVALID;
char	* cmdLineForeground = NULL;
char	* cmdLineBackground = NULL;
char	* cmdLineBorder = NULL;
int	geometrySpecified = FALSE;
int	exitNormally = FALSE;
char	geometrySpec[80];
char	* displayName = NULL;
char	initMsg1[80], initMsg2[80];

int			numExtraWins = 0;
struct WinInfoRec	* extraWinInfo = NULL;

static int	maxExtraWins = 0;
static int	quitDraw = TRUE;
static char	* xEnvStr = NULL;

int AddExtraWinInfo (win, mapped, raise, expose_handler, ev_handler,
      cleanup_routine)
   Window	win;
   int		mapped, raise;
   void		(* expose_handler)();
   int		(* ev_handler)();
   void		(* cleanup_routine)();
{
   register int	i;

   for (i = 0; i < numExtraWins; i++)
      if (extraWinInfo[i].window == None)
         break;

   if (i == numExtraWins && numExtraWins == maxExtraWins)
   {
      maxExtraWins += 10;
      extraWinInfo = (struct WinInfoRec *) realloc (extraWinInfo,
            maxExtraWins*sizeof(struct WinInfoRec));
   }
   extraWinInfo[i].window = win;
   extraWinInfo[i].mapped = mapped;
   extraWinInfo[i].raise = raise;
   extraWinInfo[i].expose_handler = expose_handler;
   extraWinInfo[i].ev_handler = ev_handler;
   extraWinInfo[i].cleanup = cleanup_routine;
   return ((i==numExtraWins) ? numExtraWins++ : i);
}

static
void InitExtraWinInfo ()
{
   extraWinInfo = (struct WinInfoRec *) calloc (10, sizeof (struct WinInfoRec));
   maxExtraWins += 10;
}

static
void CleanUpExtraWinInfo ()
{
   register int	i;

   if (extraWinInfo != NULL)
   {
      for (i = 0; i < numExtraWins; i++)
         if (extraWinInfo[i].window != None)
            (*(extraWinInfo[i].cleanup))();
      free (extraWinInfo);
   }
   numExtraWins = 0;
   maxExtraWins = 0;
   extraWinInfo = NULL;
}

static
void ExecWithFile (CmdName, FileName)
   char	* CmdName, * FileName;
{
   int	pid;
   char	s[255];

   sprintf (s, "xterm -bd red -e %s %s", CmdName, FileName);
   pid = fork ();
   if (pid == 0)
   {
      system (s);
      exit (0);
   }
}

void DeallocStrings (FStr, Str1, Menu1, Str2, Menu2, Str3, Menu3)
   char	* * FStr, * * Str1, * * Menu1, * * Str2, * * Menu2, * * Str3, * * Menu3;
{
   free (*FStr);
   free (*Str1);
   free (*Menu1);
   free (*Str2);
   free (*Menu2);
   free (*Str3);
   free (*Menu3);
}

static
void AllocStrings (FStr, Str1, Menu1, Str2, Menu2, Str3, Menu3)
   char	* * FStr, * * Str1, * * Menu1, * * Str2, * * Menu2, * * Str3, * * Menu3;
{
   char		* s;

   if((s = (char *) calloc (80, sizeof(char))) == NULL)
      Error ("AllocStrings()", "Can not calloc()");
   *FStr = s; **FStr = '\0';
   if((s = (char *) calloc (80, sizeof(char))) == NULL)
      Error ("AllocStrings()", "Can not calloc()");
   *Str1 = s; **Str1 = '\0';
   if((s = (char *) calloc (80, sizeof(char))) == NULL)
      Error ("AllocStrings()", "Can not calloc()");
   *Menu1 = s; **Menu1 = '\0';
   if((s = (char *) calloc (80, sizeof(char))) == NULL)
      Error ("AllocStrings()", "Can not calloc()");
   *Str2 = s; **Str2 = '\0';
   if((s = (char *) calloc (80, sizeof(char))) == NULL)
      Error ("AllocStrings()", "Can not calloc()");
   *Menu2 = s; **Menu2 = '\0';
   if((s = (char *) calloc (80, sizeof(char))) == NULL)
      Error ("AllocStrings()", "Can not calloc()");
   *Str3 = s; **Str3 = '\0';
   if((s = (char *) calloc (80, sizeof(char))) == NULL)
      Error ("AllocStrings()", "Can not calloc()");
   *Menu3 = s; **Menu3 = '\0';
}

/* 
 * static
 * int MyErrorHandler (display, event)
 *    Display	* display;
 *    XErrorEvent	* event;
 * {
 *    if (event->type == 0) return (TRUE);
 *    printf ("\tError: type -- %1d\n", event->type);
 *    exit (-1);
 * }
 */

void CleanUp ()
{
   TieLooseEnds ();
   CleanUpExtraWinInfo ();
   CleanUpDrawingWindow ();
   CleanUpStk ();
   CleanUpChoices ();

   CleanUpScrolls ();
   CleanUpCursors ();

   CleanUpRuler ();
   CleanUpRasters ();
   CleanUpFonts ();
   CleanUpMenu ();
   CleanUpNames ();
   CleanUpText ();
   CleanUpColors ();
   CleanUpFiles ();
   CleanUpGrids ();
   CleanUpCutBuffer ();
   CleanUpXBm ();
   CleanUpXPm ();
   CleanUpMsg ();
   CleanUpCmds ();

   if (iconWindowCreated)
   {
      XDestroyWindow (mainDisplay, iconBaseWindow);
      iconWindowCreated = FALSE;
   }
   XDestroyWindow (mainDisplay, mainWindow);
#ifndef DONT_FREE_COLORMAP
   if (newColormapUsed) XFreeColormap (mainDisplay, mainColormap);
#endif
}

void ExposeEventHandler (input, recursive)
   XEvent	* input;
   int		recursive;
{
   register int	i;
   XEvent	ev;

   if (input->xany.window == choiceWindow)
   {
      while (XCheckWindowEvent (mainDisplay, choiceWindow, ExposureMask, &ev)) ;
      RedrawChoiceWindow ();
   }
   else if (input->xany.window == drawWindow)
   {
      while (XCheckWindowEvent (mainDisplay, drawWindow, ExposureMask, &ev)) ;
      if (topSel != NULL || curChoice==VERTEXMODE || SomethingDirty ())
         ClearAndRedrawDrawWindow ();
      else
      {
         RedrawDrawWindow (botObj);
         RedrawCurText ();
      }
   }
   else if (input->xany.window == vRuleWindow)
   {
      while (XCheckWindowEvent (mainDisplay, vRuleWindow, ExposureMask, &ev)) ;
      RedrawVRuler (&input);
   }
   else if (input->xany.window == hRuleWindow)
   {
      while (XCheckWindowEvent (mainDisplay, hRuleWindow, ExposureMask, &ev)) ;
      RedrawHRuler (&input);
   }
   else if (input->xany.window == iconWindow && input->type == Expose)
   {
      while (XCheckWindowEvent (mainDisplay, iconWindow, ExposureMask, &ev)) ;
      RedrawIconWindow ();
   }
   else if (input->xany.window == titleWindow)
   {
      while (XCheckWindowEvent (mainDisplay, titleWindow, ExposureMask, &ev)) ;
      RedrawTitleWindow ();
   }
   else if (input->xany.window == msgWindow)
   {
      while (XCheckWindowEvent (mainDisplay, msgWindow, ExposureMask, &ev)) ;
      RedrawMsg ();
   }
   else if (input->xany.window == vSBarWindow)
   {
      while (XCheckWindowEvent (mainDisplay, vSBarWindow, ExposureMask, &ev)) ;
      RedrawVScrollWindow ();
   }
   else if (input->xany.window == hSBarWindow)
   {
      while (XCheckWindowEvent (mainDisplay, hSBarWindow, ExposureMask, &ev)) ;
      RedrawHScrollWindow ();
   }
   else if (pinnedMainMenu && input->xany.window == mainMenuWindow)
   {
      while (XCheckWindowEvent (mainDisplay,mainMenuWindow,ExposureMask,&ev)) ;
      RedrawMainMenuWindow ();
   }
   else if (input->xany.window==mainWindow && input->type==VisibilityNotify &&
         input->xvisibility.state==VisibilityUnobscured && !iconWindowShown)
   {
      while (XCheckWindowEvent (mainDisplay, mainWindow, VisibilityChangeMask,
            &ev)) ;
      if (pinnedMainMenu) XMapRaised (mainDisplay, mainMenuWindow);
      for (i = 0; i < numExtraWins; i++)
         if (extraWinInfo[i].window != None && extraWinInfo[i].mapped &&
               extraWinInfo[i].raise)
            XMapRaised (mainDisplay, extraWinInfo[i].window);
   }
   else
   {
      for (i = 0; i < numExtraWins; i++)
         if (input->xany.window == extraWinInfo[i].window &&
               extraWinInfo[i].window != None)
         {
            while (XCheckWindowEvent (mainDisplay, extraWinInfo[i].window,
                  ExposureMask, &ev)) ;
            (*(extraWinInfo[i].expose_handler))(input);
            break;
         }
   }

   if (recursive && (XCheckMaskEvent (mainDisplay, ExposureMask, &ev) ||
         XCheckMaskEvent (mainDisplay, VisibilityChangeMask, &ev)))
      ExposeEventHandler (&ev, recursive);
}

void MainLoop (Op, FileName, FuncStr, Str1, Menu1, Str2, Menu2, Str3, Menu3)
   char	* Op, * FileName, * * FuncStr;
   char	* * Str1, * * Menu1, * * Str2, * * Menu2, * * Str3, * * Menu3;
{
   XEvent		input;
   char			* c_ptr, file_name[MAXPATHLENGTH+1], s[MAXPATHLENGTH+1];
   char			full_name[MAXPATHLENGTH+1], * rest;
   int			rc, len, i, read_status, short_name;
   struct ObjRec	* obj_ptr;
   FILE			* fp;
   int			tmp_linenum, obj_ext_len;
   char			tmp_filename[MAXPATHLENGTH+1], obj_ext_str[MAXSTRING+1];

/* printf ("--> MainLoop (%s, %s, ...)\n", Op, FileName); */

   if (strcmp (Op, "init") == 0 || strcmp (Op, "justinit") == 0)
   {
      if (TGIF_PATCHLEVEL == 0)
         sprintf (initMsg1, "%s Version %s", TOOL_NAME, version_string);
      else
         sprintf (initMsg1, "%s Version %s (patchlevel %1d)", TOOL_NAME,
               version_string, TGIF_PATCHLEVEL);
      sprintf (initMsg2, "%s%s",
            "Copyright (C) 1990, 1991, 1992, William Chia-Wei Cheng ",
            "(william@cs.UCLA.edu)");
      fprintf (stderr, "%s\n", initMsg1);
      fprintf (stderr, "%s\n", initMsg2);
      fflush(stderr);

#ifdef USE_XAPPLRESDIR
      if ((getenv ("XENVIRONMENT") == NULL) &&
            ((c_ptr = getenv ("XAPPLRESDIR")) != NULL) &&
            ((len = strlen (c_ptr)) > 0))
      {
         FILE	* xenv_fp;

         xEnvStr = (char *) calloc (
               strlen("XENVIRONMENT=")+strlen(TOOL_NAME)+len+3, sizeof(char));
         sprintf (xEnvStr, "%s%s%s", c_ptr,
               ((c_ptr[len-1] == '/') ? "" : "/"), TOOL_NAME);
         if ((xenv_fp = fopen (xEnvStr, "r")) != NULL)
         {
            fclose (xenv_fp);
            sprintf (xEnvStr, "XENVIRONMENT=%s%s%s", c_ptr,
                  ((c_ptr[len-1] == '/') ? "" : "/"), TOOL_NAME);
            if (putenv (xEnvStr))
            {
               sprintf (s, "Warning:  Can not putenv(%s).", xEnvStr);
               Msg (s);
               *xEnvStr = '\0';
               free (xEnvStr);
               xEnvStr = NULL;
            }
         }
      }
#endif

      exitNormally = FALSE;
      if ((mainDisplay = XOpenDisplay (displayName)) == 0)
      {
         if (displayName == NULL)
         {
            fprintf (stderr, "Could not open the default display!  Abort!\n");
            exit (-1);
         }
         else
         {
            fprintf (stderr, "Could not open display '%s'!  Abort!\n",
                  displayName);
            exit (-1);
         }
      }
      mainScreen = DefaultScreen (mainDisplay);
      mainColormap = DefaultColormap (mainDisplay, mainScreen);
      mainDepth = DefaultDepth (mainDisplay, mainScreen);
      mainVisual = DefaultVisual (mainDisplay, mainScreen);
      rootWindow = RootWindow (mainDisplay, mainScreen);

      XSetErrorHandler (EmergencySaveForX);
      XSetIOErrorHandler (IOEmergencySaveForX);
      signal (SIGHUP, EmergencySave);
      signal (SIGFPE, EmergencySave);
      signal (SIGBUS, EmergencySave);
      signal (SIGSEGV, EmergencySave);

      Setup ();
      InitExtraWinInfo ();
      TwoLineMsg (initMsg1, initMsg2);
      quitDraw = FALSE;

      if (FileName[0] != '\0')
      {
         sprintf (obj_ext_str, ".%s", OBJ_FILE_EXT);
         obj_ext_len = strlen (obj_ext_str);
         len = strlen (FileName);

         if (len >= obj_ext_len &&
               strcmp (&FileName[len-obj_ext_len], obj_ext_str) == 0)
            strcpy (file_name, FileName);
         else
            sprintf (file_name, "%s.%s", FileName, OBJ_FILE_EXT);

         if (short_name = IsPrefix (bootDir, file_name, &rest)) ++rest;
         if ((fp = fopen (file_name, "r")) == NULL)
         {
            if (short_name)
               sprintf (s, "Can not open '%s'.", rest);
            else
               sprintf (s, "Can not open '%s'.", file_name);
            Msg (s);

            if (*file_name == '/')
               strcpy (full_name, file_name);
            else
               sprintf (full_name, "%s/%s", curDir, file_name);

            SetCurDir (full_name);
            *curSymDir = '\0';
            curFileDefined = TRUE;

            SetFileModified (FALSE);
            if (short_name)
               sprintf (s, "Current (empty) file is '%s'.", rest);
            else
               sprintf (s, "Current (empty) file is '%s'.", file_name);
            Msg (s);
         }
         else
         {
            if (short_name)
               sprintf (s, "Loading '%s' ...", rest);
            else
               sprintf (s, "Loading '%s' ...", file_name);
            Msg (s);

            strcpy (tmp_filename, scanFileName);
            tmp_linenum = scanLineNum;
            strcpy (scanFileName, (short_name ? rest : file_name));
            scanLineNum = 0;

            SetWatchCursor (drawWindow);
            SetWatchCursor (mainWindow);

            while ((read_status = ReadObj (fp, &obj_ptr)) == TRUE)
               if (obj_ptr != NULL)
               {
                  AdjForOldVersion (obj_ptr);
                  AddObj (NULL, topObj, obj_ptr);
               }

            strcpy (scanFileName, tmp_filename);
            scanLineNum = tmp_linenum;

            fclose (fp);

            if (read_status == INVALID)
            {
               sprintf (s, "File version too large (=%1d).  Load aborted!",
                     fileVersion);
               Msg (s);
            }
            else
            {
               if (*file_name == '/')
                  strcpy (full_name, file_name);
               else
                  sprintf (full_name, "%s/%s", curDir, file_name);

               SetCurDir (full_name);
               *curSymDir = '\0';
               curFileDefined = TRUE;

               SetFileModified (FALSE);
               if (short_name)
                  sprintf (s, "Current file is '%s'.", rest);
               else
                  sprintf (s, "Current file is '%s'.", file_name);
               Msg (s);
            }
            SetDefaultCursor (mainWindow);
            SetDefaultCursor (drawWindow);
         }
      }
      SetDefaultDrawWinClipRecs ();
      UpdateDirInfo ();
   }
   SaveDrawWinInfo ();

   if (strcmp (Op, "justinit") == 0)
   {
      XSync (mainDisplay, False);
      AllocStrings (FuncStr, Str1, Menu1, Str2, Menu2, Str3, Menu3);
      strcpy (*FuncStr, "");
      return;
   }

   if (strcmp (Op, "save") == 0)
   {
      SaveFile ();
      AllocStrings (FuncStr, Str1, Menu1, Str2, Menu2, Str3, Menu3);
      strcpy (*FuncStr, "");
      return;
   }

   if (strcmp (Op, "vi") == 0)
      ExecWithFile ("vi", FileName);

   if (strcmp (Op, "less") == 0)
      ExecWithFile ("less", FileName);

   if (strcmp (Op, "quit") == 0)
   {
      CleanUp ();
      quitDraw = TRUE;
      XFlush (mainDisplay);
      XSync (mainDisplay, True);
      XCloseDisplay (mainDisplay);
      exitNormally = TRUE;
      AllocStrings (FuncStr, Str1, Menu1, Str2, Menu2, Str3, Menu3);
      strcpy (*FuncStr, "Quit");
      strcpy (*Str1, "");
      return;
   }

   if (strcmp (Op, "msg") == 0)
   {
      Msg (FileName);
      AllocStrings (FuncStr, Str1, Menu1, Str2, Menu2, Str3, Menu3);
      strcpy (*FuncStr, "");
      return;
   }

   if (strcmp (Op, "dialog") == 0)
   {
      Dialog (FileName, "", file_name);
      AllocStrings (FuncStr, Str1, Menu1, Str2, Menu2, Str3, Menu3);
      strcpy (*FuncStr, file_name);
      strcpy (*Str1, "");
      return;
   }

   if (strcmp (Op, "mainmenu") == 0 && quitDraw)
   {
      AllocStrings (FuncStr, Str1, Menu1, Str2, Menu2, Str3, Menu3);
      strcpy (*FuncStr, "Fail");
      strcpy (*Str1, "");
      return;
   }

   while (TRUE)
   {
      exitNormally = FALSE;
      XNextEvent (mainDisplay, &input);

      if (input.type == KeyRelease)
         continue;
      else if (input.type == KeyPress)
      {
         rc = ShortHand (&input);
         switch (rc)
         {
            case BAD: /* <CONTROL> or <META> */ continue;
            case INVALID: break;
            default:
               AllocStrings (FuncStr, Str1, Menu1, Str2, Menu2, Str3, Menu3);
               strcpy (*FuncStr, fileMenuStr[rc]);
               strcpy (*Str1, curDomainName);
               sprintf (*Menu1, "tmpmodel.%s", OBJ_FILE_EXT);
               strcpy (*Str2, "");
               for (c_ptr = *FuncStr; *c_ptr != '\0'; c_ptr++)
                  if (*c_ptr == ' ')
                  {
                     *c_ptr = '\0';
                     break;
                  }
               return;
         }
      }

      if (input.xany.window == choiceWindow)
      {
         if ((rc = ChoiceEventHandler (&input)) != INVALID)
         {
            AllocStrings (FuncStr, Str1, Menu1, Str2, Menu2, Str3, Menu3);
            strcpy (*FuncStr, fileMenuStr[rc]);
            strcpy (*Str1, curDomainName);
            sprintf (*Menu1, "tmpmodel.%s", OBJ_FILE_EXT);
            strcpy (*Str2, "");
            for (c_ptr = *FuncStr; *c_ptr != '\0'; c_ptr++)
               if (*c_ptr == ' ')
               {
                  *c_ptr = '\0';
                  break;
               }
            return;
         }
      }
      else if (input.xany.window == drawWindow)
      {
         if ((rc = DrawingEventHandler (&input)) != INVALID)
         {
            AllocStrings (FuncStr, Str1, Menu1, Str2, Menu2, Str3, Menu3);
            strcpy (*FuncStr, fileMenuStr[rc]);
            strcpy (*Str1, curDomainName);
            sprintf (*Menu1, "tmpmodel.%s", OBJ_FILE_EXT);
            strcpy (*Str2, "");
            for (c_ptr = *FuncStr; *c_ptr != '\0'; c_ptr++)
               if (*c_ptr == ' ')
               {
                  *c_ptr = '\0';
                  break;
               }
            return;
         }
      }
      else if (input.xany.window == mainWindow)
      {
         if ((rc = mainWinEventHandler (&input)) != INVALID)
         {
            AllocStrings (FuncStr, Str1, Menu1, Str2, Menu2, Str3, Menu3);
            strcpy (*FuncStr, fileMenuStr[rc]);
            strcpy (*Str1, curDomainName);
            sprintf (*Menu1, "tmpmodel.%s", OBJ_FILE_EXT);
            strcpy (*Str2, "");
            for (c_ptr = *FuncStr; *c_ptr != '\0'; c_ptr++)
               if (*c_ptr == ' ')
               {
                  *c_ptr = '\0';
                  break;
               }
            return;
         }
      }
      else if (input.xany.window == vRuleWindow)
         RedrawVRuler (&input);
      else if (input.xany.window == hRuleWindow)
         RedrawHRuler (&input);
      else if (input.xany.window == iconWindow ||
            input.xany.window == iconBaseWindow)
         IconEventHandler (&input);
      else if (input.xany.window == titleWindow)
         TitleEventHandler (&input);
      else if (input.xany.window == msgWindow)
         MsgEventHandler (&input);
      else if (input.xany.window == vSBarWindow ||
            input.xany.window == hSBarWindow)
         ScrollEventHandler (&input);
      else if (input.xany.type == MappingNotify)
         XRefreshKeyboardMapping (&(input.xmapping));
      else if (pinnedMainMenu && input.xany.window == mainMenuWindow)
      {
         if ((rc = MainMenuEventHandler (&input)) != INVALID)
         {
            AllocStrings (FuncStr, Str1, Menu1, Str2, Menu2, Str3, Menu3);
            strcpy (*FuncStr, fileMenuStr[rc]);
            strcpy (*Str1, curDomainName);
            sprintf (*Menu1, "tmpmodel.%s", OBJ_FILE_EXT);
            strcpy (*Str2, "");
            for (c_ptr = *FuncStr; *c_ptr != '\0'; c_ptr++)
               if (*c_ptr == ' ')
               {
                  *c_ptr = '\0';
                  break;
               }
            return;
         }
      }
      else
      {
         for (i = 0; i < numExtraWins; i++)
            if (input.xany.window == extraWinInfo[i].window &&
                  extraWinInfo[i].window != None)
            {
               if ((rc = (*(extraWinInfo[i].ev_handler))(&input)) != INVALID)
               {
                  AllocStrings (FuncStr, Str1, Menu1, Str2, Menu2, Str3, Menu3);
                  strcpy (*FuncStr, fileMenuStr[rc]);
                  strcpy (*Str1, curDomainName);
                  sprintf (*Menu1, "tmpmodel.%s", OBJ_FILE_EXT);
                  strcpy (*Str2, "");
                  for (c_ptr = *FuncStr; *c_ptr != '\0'; c_ptr++)
                     if (*c_ptr == ' ')
                     {
                        *c_ptr = '\0';
                        break;
                     }
                  return;
               }
               break;
            }
      }
   }
}

static
void HandleSimpleEvent (input)
   XEvent	input;
{
   if ((input.type & (PointerMotionMask | EnterWindowMask | LeaveWindowMask))
         != 0) return;

   if (input.xany.window == drawWindow)
      DrawingEventHandler (&input);
   else if (input.xany.window == choiceWindow)
      ChoiceEventHandler (&input);
   else if (input.xany.window==iconWindow || input.xany.window==iconBaseWindow)
      IconEventHandler (&input);
   else if (input.xany.window == titleWindow)
      TitleEventHandler (&input);
   else if (input.xany.window == msgWindow)
      MsgEventHandler (&input);
   else if (input.xany.window==vSBarWindow || input.xany.window==hSBarWindow)
      ScrollEventHandler (&input);
   else if (input.xany.window == hRuleWindow)
      RedrawHRuler ();
   else if (input.xany.window == vRuleWindow)
      RedrawVRuler ();
   else if (pinnedMainMenu && input.xany.window == mainMenuWindow)
      MainMenuEventHandler (&input);
}

static
void AllocReturnStr (ReturnStr)
   char	* * ReturnStr;
{
   if((*ReturnStr = (char *) calloc (80, sizeof(char))) == NULL)
      Error ("AllocReturnStr()", "calloc fails");
}

void Animate (TypeStr, PolyId, SpeedStr, ColorStr, ReturnStr)
   char		* TypeStr, * PolyId, * SpeedStr, * ColorStr;
   char		* * ReturnStr;
{
   struct ObjRec	* obj_ptr;
   char			s[80];
   int			i, poly_id, speed, pixel, clicked = FALSE;
   XEvent		input;
   XButtonEvent		* button_event;

   AllocReturnStr (ReturnStr);
   strcpy (*ReturnStr, "");

   while (XPending (mainDisplay) != 0)
   {
      XPeekEvent (mainDisplay, &input);
      if ((input.type & (PointerMotionMask | ExposureMask | EnterWindowMask |
            LeaveWindowMask)) != 0)
      {
         XNextEvent (mainDisplay, &input);
         HandleSimpleEvent (input);
      }
      else
      {
         if (input.type == ButtonPress) clicked = TRUE;
         strcpy (*ReturnStr, "Interrupt");
         break;
      }
   }

   printf ("--> Animate (%s, %s, %s, %s)\n",TypeStr,PolyId,SpeedStr,ColorStr);
   if (strcmp (TypeStr, "waitclick") == 0)
   {
      Msg ("Left:step.  Middle:run.  Right:stop.");
      if (!clicked)
      {
         while(TRUE)
            if (XPending (mainDisplay) != 0)
            {
               XNextEvent (mainDisplay, &input);
               if (input.type == ButtonPress)
                  break;
            }
      }
      else
         XNextEvent (mainDisplay, &input);

      button_event = &(input.xbutton);
      switch (button_event->button)
      {
         case Button1 : strcpy (*ReturnStr, "Left"); break;
         case Button2 : strcpy (*ReturnStr, "Middle"); break;
         case Button3 : strcpy (*ReturnStr, "Right"); break;
      }
      Msg ("");
   }
   else
   {
      poly_id = atoi (PolyId);
      for (obj_ptr = topObj; obj_ptr != NULL; obj_ptr = obj_ptr->next)
         if (obj_ptr->type == OBJ_POLY && obj_ptr->id == poly_id)
            break;

      if (obj_ptr == NULL)
      {
         sprintf (s, "Can not find poly id = %1d.", poly_id);
         Msg (s);
      }
      else if (strcmp (TypeStr, "send") == 0)
      {
         speed = atoi (SpeedStr);
         pixel = (colorDisplay) ? xorColorPixels[obj_ptr->color] : 1;
         AnimateSend (obj_ptr->detail.p, speed, pixel);
      }
      else if (strcmp (TypeStr, "flash") == 0)
      {
         if (colorDisplay)
            strcpy (s, ColorStr);
         else
            strcpy (s, "white");
         for (i = 0; i < maxColors; i++)
            if (strcmp (colorMenuItems[i], s) == 0)
            {
               AnimateFlashColor (obj_ptr, i);
               break;
            }
      }
   }
}

static
void MyFormat (Str)
   char	* Str;
{
   register char	* c_ptr = Str, * period_ptr = NULL;
   register int		i;

   for ( ; *c_ptr != '\0'; c_ptr++)
      if (*c_ptr >= '0' && *c_ptr <= '9')
         continue;
      else if (*c_ptr == '.')
      {
         if (period_ptr != NULL) return;
         period_ptr = c_ptr;
         continue;
      }
      else
         return;

   if (period_ptr == NULL) return;
   for (c_ptr = period_ptr, i = 0; *c_ptr != '\0' && i < 7; c_ptr++, i++) ;
   if (*c_ptr != '\0') *c_ptr = '\0';
}

void UpdAttrVal (ObjId, AttrName, AttrColor, AttrVal, ReturnStr)
   char		* ObjId, * AttrName, * AttrColor, * AttrVal;
   char		* * ReturnStr;
{
   struct ObjRec	* obj_ptr;
   struct AttrRec	* attr_ptr;
   char			s[80];
   int			obj_id;
   XEvent		input;
   XButtonEvent		* button_event;

   AllocReturnStr (ReturnStr);
   strcpy (*ReturnStr, "");

   while (XPending (mainDisplay) != 0)
   {
      XPeekEvent (mainDisplay, &input);
      if ((input.type & (PointerMotionMask | ExposureMask | EnterWindowMask |
            LeaveWindowMask)) != 0)
      {
         XNextEvent (mainDisplay, &input);
         HandleSimpleEvent (input);
      }
      else
      {
         strcpy (*ReturnStr, "Interrupt");
         button_event = &(input.xbutton);
         switch (button_event->button)
         {
            case Button1 : strcpy (*ReturnStr, "Left"); break;
            case Button2 : strcpy (*ReturnStr, "Middle"); break;
            case Button3 : strcpy (*ReturnStr, "Right"); break;
         }
         break;
      }
   }

   printf ("--> UpdAttrVal (%s, %s, %s, %s)\n", ObjId, AttrName,
         AttrColor, AttrVal);

   obj_id = atoi (ObjId);

   for (obj_ptr = topObj; obj_ptr != NULL; obj_ptr = obj_ptr->next)
      if (obj_ptr->id == obj_id)
         break;

   if (obj_ptr == NULL)
   {
      sprintf (s, "Can not find obj id = %1d.", obj_id);
      Msg (s);
   }
   else
   {
      attr_ptr = obj_ptr->fattr;
      for ( ; attr_ptr != NULL; attr_ptr = attr_ptr->next)
      {
         if (strcmp (AttrName, attr_ptr->name) == 0 &&
               strcmp (AttrColor, colorMenuItems[attr_ptr->obj->color]) == 0)
            break;
      }
      if (attr_ptr == NULL)
      {
         sprintf (s, "Can not find attr name '%s' and color '%s'.",
               AttrName, AttrColor);
         Msg (s);
      }
      else
      {
         strcpy (attr_ptr->s, AttrVal);
         MyFormat(attr_ptr->s);
         if (attr_ptr->nameshown)
            sprintf (s, "%s%s", attr_ptr->name, attr_ptr->s);
         else
            strcpy (s, attr_ptr->s);

         if (attr_ptr->shown) RepaintFirstStr (attr_ptr->obj, s);

         strcpy (attr_ptr->obj->detail.t->first->s, s);
      }
   }
}
